<?php
/**
 * All abstract representations of DocBlock tags are defined
 * by the classes in this file
 * @package phpDocumentor
 * @subpackage DocBlockTags
 * @since separate file since version 1.2
 */
/**
 * used to represent standard tags like @access, etc.
 * This class is aware of inline tags, and will automatically handle them using inherited functions
 * @package phpDocumentor
 * @subpackage DocBlockTags
 * @author Greg Beaver <cellog@users.sourceforge.net>
 * @since 1.0rc1
 * @version $Id: DocBlockTags.inc,v 1.3 2003/01/18 07:25:19 CelloG Exp $
 */
class parserTag extends parserStringWithInlineTags
{
    /**
     * Type is used by many functions to skip the hassle of if get_class($blah) == 'parserBlah'
     * always '_tag'
     * @var string
     */
    var $type = '_tag';
    /**
     * tag name (see, access, etc.)
     * @var string
     */
    var $keyword = '';
    
    /**
     * @param string $keyword tag name
     * @param parserStringWithInlineTags $value
     */
    function parserTag($keyword, $value, $noparse = false)
    {
        $this->keyword = $keyword;
        if (!$noparse)
        {
            $parser = new parserDescParser;
            $parser->subscribe('*',$this);
            $parser->parse($value->value,true,'parserStringWithInlineTags');
        } else $this->value = $value;
    }
    
    /**
     * @param Converter
     * @see Converter
     */
    function Convert(&$converter)
    {
        if (is_array($this->value))
        {
            $result = '';
            foreach($this->value as $val)
            {
                $result .= $val->Convert($converter);
            }
            return $result;
        } else
        return $this->value->Convert($converter);
    }
    
    function HandleEvent($a,$desc)
    {
        $this->value = $desc;
    }
    
    /**
     * @return string returns the text minus any inline tags
     * @see parserStringWithInlineTags::getString()
     */
    function getString()
    {
        return $this->value->getString();
    }
}

class parserNameTag extends parserTag
{
    /**
     * tag name
     * @var string
     */
    var $keyword = 'name';
    
    function parserNameTag($name, $value)
    {
        $this->value = $value;
    }
    
    /**
     * @see parserStringWithInlineTags::Convert()
     * @param Converter
     * @return string converted value of the tag
     */
    function Convert(&$c)
    {
        return $this->value;
    }
}

class parserAccessTag extends parserTag
{
    /**
     * tag name
     * @var string
     */
    var $keyword = 'access';
    
    /**
     * set to true if the returned tag has a value type of private or public, false otherwise
     * @var boolean
     */
    var $isvalid = false;
    
    /**
     * checks $value to make sure it is private or public, otherwise it's not a valid @access tag
     * @see $isvalid
     * @param parserStringWithInlineTags $value
     */
    function parserAccessTag($value)
    {
        if (!is_string($value))
        {
            if (is_object($value))
            {
                if (method_exists($value,'getstring'))
                {
                    $value = $value->getString();
                }
            }
        }
        switch(trim($value))
        {
            case 'private' :
            case 'public' :
            case 'protected' :
                $this->value = $value;
                $this->isvalid = true;
            break;
            default :
            addError(PDERROR_ACCESS_WRONG_PARAM,$value);
                $this->value = 'public';
            break;
        }
    }
    
    /**
     * @see parserStringWithInlineTags::Convert()
     * @param Converter
     * @return string converted value of the tag
     */
    function Convert(&$converter)
    {
        return $this->value;
    }
    
    /**
     * No inline tags are possible, returns 'public' or 'private'
     * @return string returns the text minus any inline tags
     */
    function getString()
    {
        return $this->value;
    }
}

/**
 * represents "@return"
 * @author Greg Beaver <cellog@users.sourceforge.net>
 * @since 1.0rc1
 * @version $Id: DocBlockTags.inc,v 1.3 2003/01/18 07:25:19 CelloG Exp $
 */
class parserReturnTag extends parserTag
{
    /**
     * always 'return'
     * @var string
     */
    var $keyword = 'return';
    /**
     * the type a function returns
     */
    var $returnType = 'void';
    
    /**
     * contains a link to the documentation for a class passed as a type in @return, @var or @param
     * Example:
     * <code>
     * class myclass
     * {
     * ...
     * }
     * /** @return myclass blahblahblah
     * ...
     * </code>
     * In this case, $converted_returnType will contain a link to myclass instead of the string 'myclass'
     * @var mixed either the same as $returnType or a link to the docs for a class
     * @see $returnType
     */
    var $converted_returnType = false;
    
    /**
     * @param string
     * @param parserStringWithInlineTags
     */
    function parserReturnTag($returnType, $value)
    {
        $this->returnType = $returnType;
        parent::parserTag('return',$value);
    }
    
    /**
     * sets up $converted_returnType
     * @see parserStringWithInlineTags::Convert(), $converted_returnType
     * @param Converter
     * @return string converted value of the tag
     */
    function Convert(&$converter)
    {
        $my_types = '';
        if (strpos($this->returnType,'|'))
        {
            $types = explode('|',$this->returnType);
            foreach($types as $returntype)
            {
                $a = $converter->getLink($returntype);
                if (is_object($a) && get_class($a) == 'classlink')
                {
                    if (!empty($my_types)) $my_types .= '|';
                    $my_types .= $converter->returnSee($a,$converter->type_adjust($returntype));
                } else
                {
                    if (!empty($my_types)) $my_types .= '|';
                    $my_types .= $converter->type_adjust($returntype);
                }
            }
            $this->converted_returnType = $my_types;
        } else
        {
            $a = $converter->getLink($this->returnType);
            if (is_object($a) && get_class($a) == 'classlink')
            {
                $this->converted_returnType = $converter->returnSee($a,$converter->type_adjust($this->returnType));
            } else
            {
                $this->converted_returnType = $converter->type_adjust($this->returnType);
            }
        }
        return parserTag::Convert($converter);
    }
}

/**
 * represents "@var"
 * @author Greg Beaver <cellog@users.sourceforge.net>
 * @since 1.0rc1
 * @version $Id: DocBlockTags.inc,v 1.3 2003/01/18 07:25:19 CelloG Exp $
 */
class parserVarTag extends parserReturnTag
{
    /**
     * always 'var'
     * @var string
     */
    var $keyword = 'var';
    /**
     * the type a var has
     * @var string
     */
    var $returnType = 'mixed';
}

class parserParamTag extends parserVarTag
{
    /**
     * always 'param'
     * @var string
     */
    var $keyword = 'param';
}

class parserStaticvarTag extends parserParamTag
{
    /**
     * always 'staticvar'
     * @var string
     */
    var $keyword = 'staticvar';
}

/**
 * represents "@link"
 * @author Greg Beaver <cellog@users.sourceforge.net>
 * @since 1.0rc1
 * @version $Id: DocBlockTags.inc,v 1.3 2003/01/18 07:25:19 CelloG Exp $
 */
class parserLinkTag extends parserTag
{
    /**
     * always 'link'
     * @var string
     */
    var $keyword = 'link';
    
    /**
     * URL to link to
     * @param string $link
     */
    function parserLinkTag($link)
    {
        $a = new parserLinkInlineTag($link->getString(),$link->getString());
        $b = new parserStringWithInlineTags;
        $b->add($a);
        $this->value = $b;
    }
}

/**
 * represents "@see"
 * @author Greg Beaver <cellog@users.sourceforge.net>
 * @since 1.0rc1
 * @version $Id: DocBlockTags.inc,v 1.3 2003/01/18 07:25:19 CelloG Exp $
 */
class parserSeeTag extends parserLinkTag
{
    /**
     * always 'see'
     * @var string
     */
    var $keyword = 'see';
    
    function parserSeeTag($name)
    {
        parserTag::parserTag('see',$name,true);
    }

    /**
     * @param Converter
     * @see parserStringWithInlineTags::Convert()
     */
    function Convert(&$converter)
    {
        $a = $converter->getLink(trim($this->value->Convert($converter)));
        if (is_string($a))
        {
            // feature 564991
            if (strpos($a,'://'))
            {
                // php function
                return $converter->returnLink($a,str_replace('PHP_MANUAL#','',$this->value->Convert($converter)));
            }
            return $a;
        }
        if (is_object($a)) return $converter->returnSee($a);
        // getLink parsed a comma-delimited list of linked thingies, add the commas back in
        if (is_array($a))
        {
            $b = '';
            foreach($a as $i => $bub)
            {
                if (!empty($b)) $b .= ', ';
                if (is_string($a[$i])) $b .= $a[$i];
                if (is_object($a[$i])) $b .= $converter->returnSee($a[$i]);
            }
            return $b;
        }
        return false;
    }
}

/**
 * represents "@uses"
 *
 * This is exactly like @see except that the element @used has a @see link to this element added to its docblock
 * @tutorial tags.uses.pkg
 * @since 1.2
 */
class parserUsesTag extends parserSeeTag
{
    var $keyword = 'uses';
}

/**
 * This is a virtual tag, it is created by @uses to cross-reference the used element
 *
 * This is exactly like @see.
 * @since 1.2
 */
class parserUsedByTag extends parserSeeTag
{
    var $keyword = 'usedby';
    
    function parserUsedByTag($usedbylist)
    {
        $this->value = $usedbylist;
    }
    
    function Convert(&$c)
    {
        $a = '';
        foreach($this->value as $link)
        {
            if (!empty($a)) $a .= ', ';
            $a .= $c->returnSee($link);
        }
        return $a;
    }
}

/**
 * represents "@tutorial"
 *
 * This is exactly like @see except that it only links to tutorials
 * @tutorial phpDocumentor/tutorials.pkg
 * @tutorial tags.tutorial.pkg
 * @since 1.2
 */
class parserTutorialTag extends parserSeeTag
{
    var $keyword = 'tutorial';
    /**
     * @param Converter
     * @see parserStringWithInlineTags::Convert()
     */
    function Convert(&$converter)
    {
        $a = $converter->getTutorialLink(trim($this->value->Convert($converter)));
        if (is_string($a))
        {
            return $a;
        }
        if (is_object($a)) return $converter->returnSee($a);
        // getLink parsed a comma-delimited list of linked thingies, add the commas back in
        if (is_array($a))
        {
            $b = '';
            foreach($a as $i => $bub)
            {
                if (!empty($b)) $b .= ', ';
                if (is_string($a[$i])) $b .= $a[$i];
                if (is_object($a[$i])) $b .= $converter->returnSee($a[$i]);
            }
            return $b;
        }
        return false;
    }
}

/**
 * represents "@source"
 * 
 * The @source tag is unimplemented and may not be implemented
 * @tutorial tags.example.pkg
 * @todo implement support in PDF Converter
 */
class parserSourceTag extends parserTag
{
    var $keyword = 'source';
    var $source;
    var $path;
    var $writesource = 'writeSourceFragment';
    function parserSourceTag($filepath, $value)
    {
        parent::parserTag($this->keyword, '');
        $this->path = $filepath;
        $this->source = $value;
    }
    function Convert(&$c)
    {
        $this->ConvertSource($c);
        return $this->getSourceLink($c);
    }
    
    function getSourceLink(&$c)
    {
        return $c->getSourceFunctionLink($this->path,$this->_function);
    }
    
    /**
     * @todo find and link to documentation of file source
     */
    function ConvertSource(&$c)
    {
        $return = '';
        foreach($this->source as $linenum => $tokens)
        {
            $line = '';
            foreach($tokens as $token)
            { // Convert them using highlighting specific to each converter
                if (is_array($token))
                {
                    $line .= $c->highlightSource($token[0], $token[1]);
                } else
                {
                    $line .= $c->highlightSource(false, $token);
                }
            }
            $return .= $c->SourceLine($linenum + 1,rtrim($line));
        }
        $writeSource = $this->writesource;
        $this->writeSource($c, $c->PreserveWhiteSpace($return));
    }
    
    function writeSource(&$c, $source)
    {
        $c->writeSourceFunction($this->_function, $this->path, $source);
    }
}

/**
 * represents "@filesource"
 * 
 * Use this to create a link to a highlighted phpxref-style source file listing
 * @tutorial tags.filesource.pkg
 * @todo implement support in PDF Converter
 */
class parserFileSourceTag extends parserSourceTag
{
    var $keyword = 'filesource';
    var $writesource = 'writeSource';
    
    function writeSource(&$c, $source)
    {
        $c->writeSource($this->path, $source);
    }

    function getSourceLink(&$c)
    {
        return $c->getSourceLink($this->path);
    }
}

/**
 * represents "@example"
 * 
 * The example tag 
 * @tutorial tags.example.pkg
 * @todo implement support in PDF Converter
 */
class parserExampleTag extends parserSourceTag
{
    var $keyword = 'example';
    var $writesource = 'writeExample';
    function parserExampleTag($tag, $value)
    {
        parent::parserTag($tag, $value);
        $path = substr($value->getString(),0,strpos($value->getString(),".php") + 4);
        $value = substr($value->getString(),strlen($path));
        if (!$value) $value = '';
        $this->_title = $value;
        $this->path = basename($path);
        if (!@file_exists($path))
        {
            addWarning(PDERROR_EXAMPLE_NOT_FOUND,$path);
        } else
        {
            $f = @fopen($path,'r');
            if ($f)
            {
                $example = fread($f,filesize($path));
                $obj = new phpDocumentorTWordParser;
                $obj->setup($example);
                $this->source = $obj->getFileSource();
                unset($obj);
            }
        }
    }
    
    function writeSource(&$c, $source)
    {
        $c->writeExample($this->_title, $this->path, $source);
    }

    function getSourceLink(&$c)
    {
        return $c->getExampleLink($this->path, $this->_title);
    }
}

?>